"""
bilibili_api.utils.short
一个很简单的处理短链接的模块，主要是读取跳转链接。
"""
from .network_httpx import get_session
from .. import settings


async def get_real_url(short_url: str):
    """
    获取短链接跳转目标，以进行操作。
    Params:
        short_url(str): 短链接。
    Returns:
        目标链接（如果不是有效的链接会报错）
    """
    try:
        url = short_url
        headers = await get_headers(url)
        while "location" in headers.keys():
            url = headers["location"]
            headers = await get_headers(url)
        return url  # 已经是最终路径
    except Exception as e:
        raise e


async def get_headers(short_url: str):
    """
    获取链接的 headers
    """
    config = {}
    config["method"] = "GET"
    config["url"] = short_url
    config["follow_redirects"] = False
    if settings.proxy:
        config["proxies"] = {settings.proxy_use: settings.proxy}
    print(short_url)
    resp = await get_session().head(url=short_url, follow_redirects=False)
    return resp.headers



"""
IP 终端相关
"""

from .utils.network_httpx import request
from .utils.utils import get_api

API = get_api("client")


async def get_zone():
    """
    通过 IP 获取地理位置
    """
    api = API["zone"]
    return await request("GET", api["url"])


async def get_client_info():
    """
    获取 IP 信息
    """
    api = API["info"]
    return await request("GET", api["url"])



"""
表情包相关
"""

from .utils.network_httpx import request
from .utils.utils import get_api

API = get_api("emoji")


async def get_emoji_list(business: str = "reply"):
    """
    获取表情包列表

    Args:
        business(str): 使用场景, reply / dynamic
    """
    api = API["list"]
    params = {"business": business}
    return await request("GET", api["url"], params=params)



"""
bilibili_api.utils.parse_link

链接资源解析。
"""

import json
import re
from enum import Enum

import httpx

from ..article import Article, ArticleList
from ..audio import Audio, AudioList
from ..bangumi import Bangumi, Episode
from ..black_room import BlackRoom
from ..cheese import CheeseList, CheeseVideo
from ..dynamic import Dynamic
from ..exceptions import *
from ..favorite_list import FavoriteList, FavoriteListType, get_video_favorite_list
from ..interactive_video import InteractiveVideo
from ..live import LiveRoom
from ..user import ChannelSeries, ChannelSeriesType, User, get_self_info
from ..video import Video
from .Credential import Credential
from .short import get_real_url
from .sync import sync
from .utils import get_api


class ResourceType(Enum):
    """
    链接类型类。

    + VIDEO: 视频
    + BANGUMI: 番剧
    + EPISODE: 番剧剧集
    + FAVORITE_LIST: 视频收藏夹
    + CHEESE: 课程
    + CHEESE_VIDEO: 课程视频
    + AUDIO: 音频
    + AUDIO_LIST: 歌单
    + ARTICLE: 专栏
    + USER: 用户
    + LIVE: 直播间
    + CHANNEL_SERIES: 合集与列表
    + BLACK_ROOM: 小黑屋
    """

    VIDEO = "video"
    INTERACTIVE_VIDEO = "interactive_video"
    BANGUMI = "bangumi"
    EPISODE = "episode"
    FAVORITE_LIST = "favorite_list"
    CHEESE_VIDEO = "cheese_video"
    AUDIO = "audio"
    AUDIO_LIST = "audio_list"
    ARTICLE = "article"
    USER = "user"
    LIVE = "live"
    CHANNEL_SERIES = "channel_series"
    ARTICLE_LIST = "article_list"
    DYNAMIC = "dynamic"
    BLACK_ROOM = "black_room"


async def parse_link(url, credential: Credential = Credential()):
    """
    解析 bilibili url 的函数。

    Args:
        url(str)              : 链接
        credential(Credential): 凭据类

    Returns:
        Union[tuple, int]: (对象，类型) 或 -1,-1 表示出错
    """
    try:
        obj = None

        sobj = check_short_name(url, credential)
        if sobj != -1:
            sobj[0].credential = credential
            return sobj

        
        black_room = parse_black_room(url)
        if not black_room == -1:
            obj = (black_room, ResourceType.BLACK_ROOM)
            return obj

        url = await get_real_url(url)

        # 特殊处理，因为后面会过滤参数，这两项需要参数完成
        channel = parse_season_series(url)
        if channel != -1:
            return (channel, ResourceType.CHANNEL_SERIES)
        fl_space = parse_space_favorite_list(url, credential)
        if fl_space != -1:
            return fl_space

        # 过滤参数
        url = url.split("?")[0]
        if url == "https://space.bilibili.com":
            try:
                info = sync(get_self_info(credential))
            except:
                return -1
            else:
                return (User(info["mid"], credential=credential), ResourceType.USER)
        obj = None
        video = parse_video(url, credential)
        if not video == -1:
            if isinstance(video, InteractiveVideo):
                obj = (video, ResourceType.INTERACTIVE_VIDEO)
            else:
                obj = (video, ResourceType.VIDEO)
        bangumi = parse_bangumi(url)
        if not bangumi == -1:
            obj = (bangumi, ResourceType.BANGUMI)
        episode = parse_episode(url, credential)
        if not episode == -1:
            obj = (episode, ResourceType.EPISODE)
        favorite_list = parse_favorite_list(url)
        if not favorite_list == -1:
            obj = (favorite_list, ResourceType.FAVORITE_LIST)
        cheese_video = parse_cheese_video(url)
        if not cheese_video == -1:
            obj = (cheese_video, ResourceType.CHEESE_VIDEO)
        audio = parse_audio(url)
        if not audio == -1:
            obj = (audio, ResourceType.AUDIO)
        audio_list = parse_audio_list(url)
        if not audio_list == -1:
            obj = (audio_list, ResourceType.AUDIO_LIST)
        article = parse_article(url)
        if not article == -1:
            obj = (article, ResourceType.ARTICLE)
        article_list = parse_article_list(url)
        if not article_list == -1:
            obj = (article_list, ResourceType.ARTICLE_LIST)
        user = parse_user(url)
        if not user == -1:
            obj = (user, ResourceType.USER)
        live = parse_live(url)
        if not live == -1:
            obj = (live, ResourceType.LIVE)
        dynamic = parse_dynamic(url)
        if not dynamic == -1:
            obj = (dynamic, ResourceType.DYNAMIC)

        if obj == None or obj[0] == None:
            return -1
        else:
            obj[0].credential = credential
            return obj
    except Exception as e:
        raise e
        return -1


def check_short_name(name: str, credential: Credential):
    """
    解析:
      - avxxxxxxxxxx
      - bvxxxxxxxxxx
      - mlxxxxxxxxxx
      - uidxxxxxxxxx
      - cvxxxxxxxxxx
      - auxxxxxxxxxx
      - amxxxxxxxxxx
      - rlxxxxxxxxxx
    """
    if name[:2].upper() == "AV":
        v = Video(aid=int(name[2:]))
        info = json.loads(
            httpx.get(
                "https://api.bilibili.com/x/web-interface/view",
                params={"bvid": v.get_bvid()},
                cookies=credential.get_cookies(),
            ).text
        )
        is_interactive = info["data"]["rights"]["is_stein_gate"]
        if is_interactive == 1:
            return (InteractiveVideo(v.get_bvid()), ResourceType.INTERACTIVE_VIDEO)
        else:
            return (v, ResourceType.VIDEO)
    elif name[:2].upper() == "BV":
        v = Video(bvid=name)
        info = json.loads(
            httpx.get(
                "https://api.bilibili.com/x/web-interface/view",
                params={"bvid": v.get_bvid()},
                cookies=credential.get_cookies(),
            ).text
        )
        is_interactive = info["data"]["rights"]["is_stein_gate"]
        if is_interactive == 1:
            return (InteractiveVideo(v.get_bvid()), ResourceType.INTERACTIVE_VIDEO)
        else:
            return (v, ResourceType.VIDEO)
    elif name[:2].upper() == "ML":
        return (
            FavoriteList(FavoriteListType.VIDEO, int(name[2:])),
            ResourceType.FAVORITE_LIST,
        )
    elif name[:3].upper() == "UID":
        return (User(int(name[3:])), ResourceType.USER)
    elif name[:2].upper() == "CV":
        return (Article(int(name[2:])), ResourceType.ARTICLE)
    elif name[:2].upper() == "AU":
        return (Audio(int(name[2:])), ResourceType.AUDIO)
    elif name[:2].upper() == "AM":
        return (AudioList(int(name[2:])), ResourceType.AUDIO_LIST)
    elif name[:2].upper() == "RL":
        return (ArticleList(int(name[2:])), ResourceType.ARTICLE_LIST)
    else:
        return -1


def parse_video(url, credential: Credential):
    """
    解析视频,如果不是返回 -1，否则返回对应类
    """
    if url[:31] == "https://www.bilibili.com/video/":
        last_part = url[31:]
        if last_part[:2].upper() == "AV":
            aid = int(last_part[2:].replace("/", ""))
            v = Video(aid=aid)
        elif last_part[:2].upper() == "BV":
            bvid = "BV" + last_part[2:].replace("/", "")
            v = Video(bvid=bvid)
        else:
            return -1
        info = json.loads(
            httpx.get(
                "https://api.bilibili.com/x/web-interface/view",
                params={"bvid": v.get_bvid()},
                cookies=credential.get_cookies(),
            ).text
        )
        is_interactive = info["data"]["rights"]["is_stein_gate"]
        if is_interactive == 1:
            return InteractiveVideo(v.get_bvid())
        else:
            return v
    else:
        return -1


def parse_bangumi(url):
    """
    解析番剧,如果不是返回 -1，否则返回对应类
    """
    if url[:41] == "https://www.bilibili.com/bangumi/media/md":
        last_part = url[41:].replace("/", "")
        media_id = int(last_part)
        return Bangumi(media_id=media_id)
    else:
        return -1


def parse_episode(url, credential):
    """
    解析番剧剧集,如果不是返回 -1，否则返回对应类
    """
    if url[:38] == "https://www.bilibili.com/bangumi/play/":
        last_part = url[38:]
        if last_part[:2].upper() == "EP":
            epid = int(last_part[2:].replace("/", ""))
            return Episode(epid=epid)
        elif last_part[:2].upper() == "SS":
            try:
                resp = httpx.get(
                    url,
                    cookies=credential.get_cookies(),
                    headers={"User-Agent": "Mozilla/5.0"},
                )
            except Exception as e:
                raise ResponseException(str(e))
            else:
                content = resp.text

                pattern = re.compile(r"window.__INITIAL_STATE__=(\{.*?\});")
                match = re.search(pattern, content)
                if match is None:
                    raise ApiException("未找到番剧信息")
                try:
                    content = json.loads(match.group(1))
                except json.JSONDecodeError:
                    raise ApiException("信息解析错误")
                else:
                    epid = content["epInfo"]["id"]
                    return Episode(epid=epid)
        else:
            return -1
    else:
        return -1


def parse_favorite_list(url):
    """
    解析收藏夹,如果不是返回 -1，否则返回对应类
    """
    if url[:44] == "https://www.bilibili.com/medialist/detail/ml":
        last_part = int(url[44:].replace("/", ""))
        return FavoriteList(media_id=last_part)
    else:
        return -1


def parse_cheese_video(url):
    """
    解析课程视频,如果不是返回 -1，否则返回对应类
    """
    if url[:37] == "https://www.bilibili.com/cheese/play/":
        if url[37:39].upper() == "EP":
            last_part = int(url[39:].replace("/", ""))
            return CheeseVideo(epid=last_part)
    else:
        return -1


def parse_audio(url):
    """
    解析音频,如果不是返回 -1，否则返回对应类
    """
    if url[:33] == "https://www.bilibili.com/audio/au":
        last_part = int(url[33:].replace("/", ""))
        return Audio(auid=last_part)
    else:
        return -1


def parse_audio_list(url):
    """
    解析歌单,如果不是返回 -1，否则返回对应类
    """
    if url[:33] == "https://www.bilibili.com/audio/am":
        last_part = int(url[33:].replace("/", ""))
        return AudioList(amid=last_part)
    else:
        return -1


def parse_article(url):
    """
    解析专栏,如果不是返回 -1，否则返回对应类
    """
    if url[:32] == "https://www.bilibili.com/read/cv":
        last_part = int(url[32:].replace("/", ""))
        return Article(cvid=last_part)
    else:
        return -1


def parse_user(url):
    if url[:27] == "https://space.bilibili.com/":
        num_re = re.compile(r"\d+")
        uid = num_re.findall(url)[0]
        return User(uid=uid)
    else:
        return -1


def parse_live(url):
    if url[:26] == "https://live.bilibili.com/":
        last_part = int(url[26:].replace("/", ""))
        return LiveRoom(room_display_id=last_part)
    else:
        return -1


def parse_season_series(url):
    if url[:27] == "https://space.bilibili.com/":
        uid = 0
        for i in url.split("/"):
            try:
                uid = int(i)
            except:
                pass
            if "collectiondetail" in i:
                sid = int(i[21:])
                return ChannelSeries(uid, ChannelSeriesType.SEASON, id_=sid)
            if "seriesdetail" in i:
                sid = int(i[17:])
                return ChannelSeries(uid, ChannelSeriesType.SERIES, id_=sid)
    elif url[:40] == "https://www.bilibili.com/medialist/play/":
        for i in url.split("/"):
            if "?" in i:
                uid = int(i.split("?")[0])
                params = i.split("?")[1].split("&")
                for param in params:
                    if "business_id" in param:
                        sid = int(param[12:])
                        return ChannelSeries(uid, ChannelSeriesType.SERIES, id_=sid)
    return -1


def parse_space_favorite_list(url, credential):
    if url[:27] == "https://space.bilibili.com/":
        uid = 0
        for i in url.split("/"):
            try:
                uid = int(i)
            except:
                pass
            if "favlist" in i:
                if len(i) == len("favlist"):
                    api = get_api("favorite-list")["info"]["list_list"]
                    params = {"up_mid": uid, "type": 2}

                    favorite_lists = json.loads(
                        httpx.get(
                            api["url"], params=params, cookies=credential.get_cookies()
                        ).text
                    )["data"]

                    if favorite_lists == None:
                        return -1
                    else:
                        default_favorite_list = favorite_lists["list"][0]
                        return (
                            FavoriteList(media_id=default_favorite_list["id"]),
                            ResourceType.FAVORITE_LIST,
                        )
                oid = ""
                type_ = ""
                for arg in i.split("&"):
                    if "?" in arg:
                        arg = arg.split("?")[1]
                    if "fid" in arg:
                        oid = arg[4:]
                    if "ctype" in arg:
                        type_ = int(arg[6:])
                oid_is_number = True
                try:
                    oid_int = int(oid)
                except:
                    oid_is_number = False
                if type_ == "" and oid_is_number:
                    # 我的视频收藏夹
                    oid_int = int(oid)
                    return (FavoriteList(media_id=oid_int), ResourceType.FAVORITE_LIST)
                elif type_ != "":
                    # 我的订阅
                    if type_ == 11:
                        # 收藏的收藏夹
                        oid_int = int(oid)
                        return (
                            FavoriteList(media_id=oid_int),
                            ResourceType.FAVORITE_LIST,
                        )
                    else:
                        return -1
                elif not oid_is_number:
                    # 其他类型的收藏夹
                    if oid == FavoriteListType.ARTICLE.value:
                        return (
                            FavoriteList(
                                FavoriteListType.ARTICLE, credential=credential
                            ),
                            ResourceType.FAVORITE_LIST,
                        )
                    elif oid == FavoriteListType.CHEESE.value:
                        return (
                            FavoriteList(
                                FavoriteListType.CHEESE, credential=credential
                            ),
                            ResourceType.FAVORITE_LIST,
                        )
                else:
                    return -1
            else:
                pass
    else:
        return -1
    return -1


def parse_article_list(url):
    if url[:41] == "https://www.bilibili.com/read/readlist/rl":
        last_part = int(url[41:].replace("/", ""))
        return ArticleList(last_part)
    else:
        return -1


def parse_dynamic(url):
    if url[:23] == "https://t.bilibili.com/":
        last_part = url[23:].replace("/", "")
        if last_part == "":
            return -1
        else:
            return Dynamic(int(last_part))
    else:
        return -1


def parse_black_room(url: str):
    if url.lstrip("https:") == url:
        url = "https:" + url
    if url[:39] == 'https://www.bilibili.com/blackroom/ban/':
        last_part = url[39:].replace("/", "")
        if last_part == "":
            return -1
        else:
            return BlackRoom(int(last_part))
    else:
        return -1



#!/usr/bin/env python3

# The original author of this program, Danmaku2ASS, is StarBrilliant.
# This file is released under General Public License version 3.
# You should have received a copy of General Public License text alongside with
# this program. If not, you can obtain it at http://gnu.org/copyleft/gpl.html .
# This program comes with no warranty, the author will not be resopnsible for
# any damage or problems caused by this program.

# You can obtain a latest copy of Danmaku2ASS at:
#   https://github.com/m13253/danmaku2ass
# Please update to the latest version before complaining.

# pylint: skip-file

import argparse
import calendar
import gettext
import io
import json
import logging
import math
import os
import random
import re
import sys
import time
import xml.dom.minidom


if sys.version_info < (3,):
    raise RuntimeError("at least Python 3.0 is required")

gettext.install(
    "danmaku2ass",
    os.path.join(
        os.path.dirname(os.path.abspath(os.path.realpath(sys.argv[0] or "locale"))),
        "locale",
    ),
)


def SeekZero(function):
    def decorated_function(file_):
        file_.seek(0)
        try:
            return function(file_)
        finally:
            file_.seek(0)

    return decorated_function


def EOFAsNone(function):
    def decorated_function(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except EOFError:
            return None

    return decorated_function


@SeekZero
@EOFAsNone
def ProbeCommentFormat(f):
    tmp = f.read(1)
    if tmp == "[":
        return "Acfun"
        # It is unwise to wrap a JSON object in an array!
        # See this: http://haacked.com/archive/2008/11/20/anatomy-of-a-subtle-json-vulnerability.aspx/
        # Do never follow what Acfun developers did!
    elif tmp == "{":
        tmp = f.read(14)
        if tmp == '"status_code":':
            return "Tudou"
        elif tmp.strip().startswith('"result'):
            return "Tudou2"
    elif tmp == "<":
        tmp = f.read(1)
        if tmp == "?":
            tmp = f.read(38)
            if tmp == 'xml version="1.0" encoding="UTF-8"?><p':
                return "Niconico"
            elif tmp == 'xml version="1.0" encoding="UTF-8"?><i':
                return "Bilibili"
            elif tmp == 'xml version="2.0" encoding="UTF-8"?><i':
                return "Bilibili2"
            elif tmp == 'xml version="1.0" encoding="utf-8"?><i':
                return "Bilibili"  # tucao.cc, with the same file format as Bilibili
            elif tmp == 'xml version="1.0" encoding="Utf-8"?>\n<':
                return "Bilibili"  # Komica, with the same file format as Bilibili
            elif tmp == 'xml version="1.0" encoding="UTF-8"?>\n<':
                tmp = f.read(20)
                if tmp == "!-- BoonSutazioData=":
                    return "Niconico"  # Niconico videos downloaded with NicoFox
                else:
                    return "MioMio"
        elif tmp == "p":
            return "Niconico"  # Himawari Douga, with the same file format as Niconico Douga


#
# ReadComments**** protocol
#
# Input:
#     f:         Input file
#     fontsize:  Default font size
#
# Output:
#     yield a tuple:
#         (timeline, timestamp, no, comment, pos, color, size, height, width)
#     timeline:  The position when the comment is replayed
#     timestamp: The UNIX timestamp when the comment is submitted
#     no:        A sequence of 1, 2, 3, ..., used for sorting
#     comment:   The content of the comment
#     pos:       0 for regular moving comment,
#                1 for bottom centered comment,
#                2 for top centered comment,
#                3 for reversed moving comment
#     color:     Font color represented in 0xRRGGBB,
#                e.g. 0xffffff for white
#     size:      Font size
#     height:    The estimated height in pixels
#                i.e. (comment.count('\n')+1)*size
#     width:     The estimated width in pixels
#                i.e. CalculateLength(comment)*size
#
# After implementing ReadComments****, make sure to update ProbeCommentFormat
# and CommentFormatMap.
#


def ReadCommentsNiconico(f, fontsize):
    NiconicoColorMap = {
        "red": 0xFF0000,
        "pink": 0xFF8080,
        "orange": 0xFFCC00,
        "yellow": 0xFFFF00,
        "green": 0x00FF00,
        "cyan": 0x00FFFF,
        "blue": 0x0000FF,
        "purple": 0xC000FF,
        "black": 0x000000,
        "niconicowhite": 0xCCCC99,
        "white2": 0xCCCC99,
        "truered": 0xCC0033,
        "red2": 0xCC0033,
        "passionorange": 0xFF6600,
        "orange2": 0xFF6600,
        "madyellow": 0x999900,
        "yellow2": 0x999900,
        "elementalgreen": 0x00CC66,
        "green2": 0x00CC66,
        "marineblue": 0x33FFCC,
        "blue2": 0x33FFCC,
        "nobleviolet": 0x6633CC,
        "purple2": 0x6633CC,
    }
    dom = xml.dom.minidom.parse(f)
    comment_element = dom.getElementsByTagName("chat")
    for comment in comment_element:
        try:
            c = str(comment.childNodes[0].wholeText)
            if c.startswith("/"):
                continue  # ignore advanced comments
            pos = 0
            color = 0xFFFFFF
            size = fontsize
            for mailstyle in str(comment.getAttribute("mail")).split():
                if mailstyle == "ue":
                    pos = 1
                elif mailstyle == "shita":
                    pos = 2
                elif mailstyle == "big":
                    size = fontsize * 1.44
                elif mailstyle == "small":
                    size = fontsize * 0.64
                elif mailstyle in NiconicoColorMap:
                    color = NiconicoColorMap[mailstyle]
            yield (
                max(int(comment.getAttribute("vpos")), 0) * 0.01,
                int(comment.getAttribute("date")),
                int(comment.getAttribute("no")),
                c,
                pos,
                color,
                size,
                (c.count("\n") + 1) * size,
                CalculateLength(c) * size,
            )
        except (AssertionError, AttributeError, IndexError, TypeError, ValueError):
            logging.warning(_("Invalid comment: %s") % comment.toxml())
            continue


def ReadCommentsAcfun(f, fontsize):
    # comment_element = json.load(f)
    # after load acfun comment json file as python list, flatten the list
    # comment_element = [c for sublist in comment_element for c in sublist]
    comment_elements = json.load(f)
    comment_element = comment_elements[2]
    for i, comment in enumerate(comment_element):
        try:
            p = str(comment["c"]).split(",")
            assert len(p) >= 6
            assert p[2] in ("1", "2", "4", "5", "7")
            size = int(p[3]) * fontsize / 25.0
            if p[2] != "7":
                c = str(comment["m"]).replace("\\r", "\n").replace("\r", "\n")
                yield (
                    float(p[0]),
                    int(p[5]),
                    i,
                    c,
                    {"1": 0, "2": 0, "4": 2, "5": 1}[p[2]],
                    int(p[1]),
                    size,
                    (c.count("\n") + 1) * size,
                    CalculateLength(c) * size,
                )
            else:
                c = dict(json.loads(comment["m"]))
                yield (float(p[0]), int(p[5]), i, c, "acfunpos", int(p[1]), size, 0, 0)
        except (AssertionError, AttributeError, IndexError, TypeError, ValueError):
            logging.warning(_("Invalid comment: %r") % comment)
            continue


def ReadCommentsBilibili(f, fontsize):
    dom = xml.dom.minidom.parse(f)
    comment_element = dom.getElementsByTagName("d")
    for i, comment in enumerate(comment_element):
        try:
            p = str(comment.getAttribute("p")).split(",")
            assert len(p) >= 5
            assert p[1] in ("1", "4", "5", "6", "7", "8")
            if comment.childNodes.length > 0:
                if p[1] in ("1", "4", "5", "6"):
                    c = str(comment.childNodes[0].wholeText).replace("/n", "\n")
                    size = int(p[2]) * fontsize / 25.0
                    yield (
                        float(p[0]),
                        int(p[4]),
                        i,
                        c,
                        {"1": 0, "4": 2, "5": 1, "6": 3}[p[1]],
                        int(p[3]),
                        size,
                        (c.count("\n") + 1) * size,
                        CalculateLength(c) * size,
                    )
                elif p[1] == "7":  # positioned comment
                    c = str(comment.childNodes[0].wholeText)
                    yield (
                        float(p[0]),
                        int(p[4]),
                        i,
                        c,
                        "bilipos",
                        int(p[3]),
                        int(p[2]),
                        0,
                        0,
                    )
                elif p[1] == "8":
                    pass  # ignore scripted comment
        except (AssertionError, AttributeError, IndexError, TypeError, ValueError):
            logging.warning(_("Invalid comment: %s") % comment.toxml())
            continue


def ReadCommentsBilibili2(f, fontsize):
    dom = xml.dom.minidom.parse(f)
    comment_element = dom.getElementsByTagName("d")
    for i, comment in enumerate(comment_element):
        try:
            p = str(comment.getAttribute("p")).split(",")
            assert len(p) >= 7
            assert p[3] in ("1", "4", "5", "6", "7", "8")
            if comment.childNodes.length > 0:
                time = float(p[2]) / 1000.0
                if p[3] in ("1", "4", "5", "6"):
                    c = str(comment.childNodes[0].wholeText).replace("/n", "\n")
                    size = int(p[4]) * fontsize / 25.0
                    yield (
                        time,
                        int(p[6]),
                        i,
                        c,
                        {"1": 0, "4": 2, "5": 1, "6": 3}[p[3]],
                        int(p[5]),
                        size,
                        (c.count("\n") + 1) * size,
                        CalculateLength(c) * size,
                    )
                elif p[3] == "7":  # positioned comment
                    c = str(comment.childNodes[0].wholeText)
                    yield (time, int(p[6]), i, c, "bilipos", int(p[5]), int(p[4]), 0, 0)
                elif p[3] == "8":
                    pass  # ignore scripted comment
        except (AssertionError, AttributeError, IndexError, TypeError, ValueError):
            logging.warning(_("Invalid comment: %s") % comment.toxml())
            continue


def ReadCommentsTudou(f, fontsize):
    comment_element = json.load(f)
    for i, comment in enumerate(comment_element["comment_list"]):
        try:
            assert comment["pos"] in (3, 4, 6)
            c = str(comment["data"])
            assert comment["size"] in (0, 1, 2)
            size = {0: 0.64, 1: 1, 2: 1.44}[comment["size"]] * fontsize
            yield (
                int(comment["replay_time"] * 0.001),
                int(comment["commit_time"]),
                i,
                c,
                {3: 0, 4: 2, 6: 1}[comment["pos"]],
                int(comment["color"]),
                size,
                (c.count("\n") + 1) * size,
                CalculateLength(c) * size,
            )
        except (AssertionError, AttributeError, IndexError, TypeError, ValueError):
            logging.warning(_("Invalid comment: %r") % comment)
            continue


def ReadCommentsTudou2(f, fontsize):
    comment_element = json.load(f)
    for i, comment in enumerate(comment_element["result"]):
        try:
            c = str(comment["content"])
            prop = json.loads(str(comment["propertis"]) or "{}")
            size = int(prop.get("size", 1))
            assert size in (0, 1, 2)
            size = {0: 0.64, 1: 1, 2: 1.44}[size] * fontsize
            pos = int(prop.get("pos", 3))
            assert pos in (0, 3, 4, 6)
            yield (
                int(comment["playat"] * 0.001),
                int(comment["createtime"] * 0.001),
                i,
                c,
                {0: 0, 3: 0, 4: 2, 6: 1}[pos],
                int(prop.get("color", 0xFFFFFF)),
                size,
                (c.count("\n") + 1) * size,
                CalculateLength(c) * size,
            )
        except (AssertionError, AttributeError, IndexError, TypeError, ValueError):
            logging.warning(_("Invalid comment: %r") % comment)
            continue


def ReadCommentsMioMio(f, fontsize):
    NiconicoColorMap = {
        "red": 0xFF0000,
        "pink": 0xFF8080,
        "orange": 0xFFC000,
        "yellow": 0xFFFF00,
        "green": 0x00FF00,
        "cyan": 0x00FFFF,
        "blue": 0x0000FF,
        "purple": 0xC000FF,
        "black": 0x000000,
    }
    dom = xml.dom.minidom.parse(f)
    comment_element = dom.getElementsByTagName("data")
    for i, comment in enumerate(comment_element):
        try:
            message = comment.getElementsByTagName("message")[0]
            c = str(message.childNodes[0].wholeText)
            pos = 0
            size = int(message.getAttribute("fontsize")) * fontsize / 25.0
            yield (
                float(
                    comment.getElementsByTagName("playTime")[0].childNodes[0].wholeText
                ),
                int(
                    calendar.timegm(
                        time.strptime(
                            comment.getElementsByTagName("times")[0]
                            .childNodes[0]
                            .wholeText,
                            "%Y-%m-%d %H:%M:%S",
                        )
                    )
                )
                - 28800,
                i,
                c,
                {"1": 0, "4": 2, "5": 1}[message.getAttribute("mode")],
                int(message.getAttribute("color")),
                size,
                (c.count("\n") + 1) * size,
                CalculateLength(c) * size,
            )
        except (AssertionError, AttributeError, IndexError, TypeError, ValueError):
            logging.warning(_("Invalid comment: %s") % comment.toxml())
            continue


CommentFormatMap = {
    "Niconico": ReadCommentsNiconico,
    "Acfun": ReadCommentsAcfun,
    "Bilibili": ReadCommentsBilibili,
    "Bilibili2": ReadCommentsBilibili2,
    "Tudou": ReadCommentsTudou,
    "Tudou2": ReadCommentsTudou2,
    "MioMio": ReadCommentsMioMio,
}


def WriteCommentBilibiliPositioned(f, c, width, height, styleid):
    # BiliPlayerSize = (512, 384)  # Bilibili player version 2010
    # BiliPlayerSize = (540, 384)  # Bilibili player version 2012
    BiliPlayerSize = (672, 438)  # Bilibili player version 2014
    ZoomFactor = GetZoomFactor(BiliPlayerSize, (width, height))

    def GetPosition(InputPos, isHeight):
        isHeight = int(isHeight)  # True -> 1
        if isinstance(InputPos, int):
            return ZoomFactor[0] * InputPos + ZoomFactor[isHeight + 1]
        elif isinstance(InputPos, float):
            if InputPos > 1:
                return ZoomFactor[0] * InputPos + ZoomFactor[isHeight + 1]
            else:
                return (
                    BiliPlayerSize[isHeight] * ZoomFactor[0] * InputPos
                    + ZoomFactor[isHeight + 1]
                )
        else:
            try:
                InputPos = int(InputPos)
            except ValueError:
                InputPos = float(InputPos)
            return GetPosition(InputPos, isHeight)

    try:
        comment_args = safe_list(json.loads(c[3]))
        text = ASSEscape(str(comment_args[4]).replace("/n", "\n"))
        from_x = comment_args.get(0, 0)
        from_y = comment_args.get(1, 0)
        to_x = comment_args.get(7, from_x)
        to_y = comment_args.get(8, from_y)
        from_x = GetPosition(from_x, False)
        from_y = GetPosition(from_y, True)
        to_x = GetPosition(to_x, False)
        to_y = GetPosition(to_y, True)
        alpha = safe_list(str(comment_args.get(2, "1")).split("-"))
        from_alpha = float(alpha.get(0, 1))
        to_alpha = float(alpha.get(1, from_alpha))
        from_alpha = 255 - round(from_alpha * 255)
        to_alpha = 255 - round(to_alpha * 255)
        rotate_z = int(comment_args.get(5, 0))
        rotate_y = int(comment_args.get(6, 0))
        lifetime = float(comment_args.get(3, 4500))
        duration = int(comment_args.get(9, lifetime * 1000))
        delay = int(comment_args.get(10, 0))
        fontface = comment_args.get(12)
        isborder = comment_args.get(11, "true")
        from_rotarg = ConvertFlashRotation(
            rotate_y, rotate_z, from_x, from_y, width, height
        )
        to_rotarg = ConvertFlashRotation(rotate_y, rotate_z, to_x, to_y, width, height)
        styles = ["\\org(%d, %d)" % (width / 2, height / 2)]
        if from_rotarg[0:2] == to_rotarg[0:2]:
            styles.append("\\pos(%.0f, %.0f)" % (from_rotarg[0:2]))
        else:
            styles.append(
                "\\move(%.0f, %.0f, %.0f, %.0f, %.0f, %.0f)"
                % (from_rotarg[0:2] + to_rotarg[0:2] + (delay, delay + duration))
            )
        styles.append(
            "\\frx%.0f\\fry%.0f\\frz%.0f\\fscx%.0f\\fscy%.0f" % (from_rotarg[2:7])
        )
        if (from_x, from_y) != (to_x, to_y):
            styles.append("\\t(%d, %d, " % (delay, delay + duration))
            styles.append(
                "\\frx%.0f\\fry%.0f\\frz%.0f\\fscx%.0f\\fscy%.0f" % (to_rotarg[2:7])
            )
            styles.append(")")
        if fontface:
            styles.append("\\fn%s" % ASSEscape(fontface))
        styles.append("\\fs%.0f" % (c[6] * ZoomFactor[0]))
        if c[5] != 0xFFFFFF:
            styles.append("\\c&H%s&" % ConvertColor(c[5]))
            if c[5] == 0x000000:
                styles.append("\\3c&HFFFFFF&")
        if from_alpha == to_alpha:
            styles.append("\\alpha&H%02X" % from_alpha)
        elif (from_alpha, to_alpha) == (255, 0):
            styles.append("\\fad(%.0f,0)" % (lifetime * 1000))
        elif (from_alpha, to_alpha) == (0, 255):
            styles.append("\\fad(0, %.0f)" % (lifetime * 1000))
        else:
            styles.append(
                "\\fade(%(from_alpha)d, %(to_alpha)d, %(to_alpha)d, 0, %(end_time).0f, %(end_time).0f, %(end_time).0f)"
                % {
                    "from_alpha": from_alpha,
                    "to_alpha": to_alpha,
                    "end_time": lifetime * 1000,
                }
            )
        if isborder == "false":
            styles.append("\\bord0")
        f.write(
            "Dialogue: -1,%(start)s,%(end)s,%(styleid)s,,0,0,0,,{%(styles)s}%(text)s\n"
            % {
                "start": ConvertTimestamp(c[0]),
                "end": ConvertTimestamp(c[0] + lifetime),
                "styles": "".join(styles),
                "text": text,
                "styleid": styleid,
            }
        )
    except (IndexError, ValueError) as e:
        try:
            logging.warning(_("Invalid comment: %r") % c[3])
        except IndexError:
            logging.warning(_("Invalid comment: %r") % c)


def WriteCommentAcfunPositioned(f, c, width, height, styleid):
    AcfunPlayerSize = (560, 400)
    ZoomFactor = GetZoomFactor(AcfunPlayerSize, (width, height))

    def GetPosition(InputPos, isHeight):
        isHeight = int(isHeight)  # True -> 1
        return (
            AcfunPlayerSize[isHeight] * ZoomFactor[0] * InputPos * 0.001
            + ZoomFactor[isHeight + 1]
        )

    def GetTransformStyles(
        x=None,
        y=None,
        scale_x=None,
        scale_y=None,
        rotate_z=None,
        rotate_y=None,
        color=None,
        alpha=None,
    ):
        styles = []
        out_x, out_y = x, y
        if rotate_z is not None and rotate_y is not None:
            assert x is not None
            assert y is not None
            rotarg = ConvertFlashRotation(rotate_y, rotate_z, x, y, width, height)
            out_x, out_y = rotarg[0:2]
            if scale_x is None:
                scale_x = 1
            if scale_y is None:
                scale_y = 1
            styles.append(
                "\\frx%.0f\\fry%.0f\\frz%.0f\\fscx%.0f\\fscy%.0f"
                % (rotarg[2:5] + (rotarg[5] * scale_x, rotarg[6] * scale_y))
            )
        else:
            if scale_x is not None:
                styles.append("\\fscx%.0f" % (scale_x * 100))
            if scale_y is not None:
                styles.append("\\fscy%.0f" % (scale_y * 100))
        if color is not None:
            styles.append("\\c&H%s&" % ConvertColor(color))
            if color == 0x000000:
                styles.append("\\3c&HFFFFFF&")
        if alpha is not None:
            alpha = 255 - round(alpha * 255)
            styles.append("\\alpha&H%02X" % alpha)
        return out_x, out_y, styles

    def FlushCommentLine(f, text, styles, start_time, end_time, styleid):
        if end_time > start_time:
            f.write(
                "Dialogue: -1,%(start)s,%(end)s,%(styleid)s,,0,0,0,,{%(styles)s}%(text)s\n"
                % {
                    "start": ConvertTimestamp(start_time),
                    "end": ConvertTimestamp(end_time),
                    "styles": "".join(styles),
                    "text": text,
                    "styleid": styleid,
                }
            )

    try:
        comment_args = c[3]
        text = ASSEscape(str(comment_args["n"]).replace("\r", "\n"))
        common_styles = ["\org(%d, %d)" % (width / 2, height / 2)]
        anchor = {0: 7, 1: 8, 2: 9, 3: 4, 4: 5, 5: 6, 6: 1, 7: 2, 8: 3}.get(
            comment_args.get("c", 0), 7
        )
        if anchor != 7:
            common_styles.append("\\an%s" % anchor)
        font = comment_args.get("w")
        if font:
            font = dict(font)
            fontface = font.get("f")
            if fontface:
                common_styles.append("\\fn%s" % ASSEscape(str(fontface)))
            fontbold = bool(font.get("b"))
            if fontbold:
                common_styles.append("\\b1")
        common_styles.append("\\fs%.0f" % (c[6] * ZoomFactor[0]))
        isborder = bool(comment_args.get("b", True))
        if not isborder:
            common_styles.append("\\bord0")
        to_pos = dict(comment_args.get("p", {"x": 0, "y": 0}))
        to_x = round(GetPosition(int(to_pos.get("x", 0)), False))
        to_y = round(GetPosition(int(to_pos.get("y", 0)), True))
        to_scale_x = float(comment_args.get("e", 1.0))
        to_scale_y = float(comment_args.get("f", 1.0))
        to_rotate_z = float(comment_args.get("r", 0.0))
        to_rotate_y = float(comment_args.get("k", 0.0))
        to_color = c[5]
        to_alpha = float(comment_args.get("a", 1.0))
        from_time = float(comment_args.get("t", 0.0))
        action_time = float(comment_args.get("l", 3.0))
        actions = list(comment_args.get("z", []))
        to_out_x, to_out_y, transform_styles = GetTransformStyles(
            to_x,
            to_y,
            to_scale_x,
            to_scale_y,
            to_rotate_z,
            to_rotate_y,
            to_color,
            to_alpha,
        )
        FlushCommentLine(
            f,
            text,
            common_styles
            + ["\\pos(%.0f, %.0f)" % (to_out_x, to_out_y)]
            + transform_styles,
            c[0] + from_time,
            c[0] + from_time + action_time,
            styleid,
        )
        action_styles = transform_styles
        for action in actions:
            action = dict(action)
            from_x, from_y = to_x, to_y
            from_out_x, from_out_y = to_out_x, to_out_y
            from_scale_x, from_scale_y = to_scale_x, to_scale_y
            from_rotate_z, from_rotate_y = to_rotate_z, to_rotate_y
            from_color, from_alpha = to_color, to_alpha
            transform_styles, action_styles = action_styles, []
            from_time += action_time
            action_time = float(action.get("l", 0.0))
            if "x" in action:
                to_x = round(GetPosition(int(action["x"]), False))
            if "y" in action:
                to_y = round(GetPosition(int(action["y"]), True))
            if "f" in action:
                to_scale_x = float(action["f"])
            if "g" in action:
                to_scale_y = float(action["g"])
            if "c" in action:
                to_color = int(action["c"])
            if "t" in action:
                to_alpha = float(action["t"])
            if "d" in action:
                to_rotate_z = float(action["d"])
            if "e" in action:
                to_rotate_y = float(action["e"])
            to_out_x, to_out_y, action_styles = GetTransformStyles(
                to_x,
                to_y,
                from_scale_x,
                from_scale_y,
                to_rotate_z,
                to_rotate_y,
                from_color,
                from_alpha,
            )
            if (from_out_x, from_out_y) == (to_out_x, to_out_y):
                pos_style = "\\pos(%.0f, %.0f)" % (to_out_x, to_out_y)
            else:
                pos_style = "\\move(%.0f, %.0f, %.0f, %.0f)" % (
                    from_out_x,
                    from_out_y,
                    to_out_x,
                    to_out_y,
                )
            styles = common_styles + transform_styles
            styles.append(pos_style)
            if action_styles:
                styles.append("\\t(%s)" % ("".join(action_styles)))
            FlushCommentLine(
                f,
                text,
                styles,
                c[0] + from_time,
                c[0] + from_time + action_time,
                styleid,
            )
    except (IndexError, ValueError) as e:
        logging.warning(_("Invalid comment: %r") % c[3])


# Result: (f, dx, dy)
# To convert: NewX = f*x+dx, NewY = f*y+dy
def GetZoomFactor(SourceSize, TargetSize):
    try:
        if (SourceSize, TargetSize) == GetZoomFactor.Cached_Size:
            return GetZoomFactor.Cached_Result
    except AttributeError:
        pass
    GetZoomFactor.Cached_Size = (SourceSize, TargetSize)
    try:
        SourceAspect = SourceSize[0] / SourceSize[1]
        TargetAspect = TargetSize[0] / TargetSize[1]
        if TargetAspect < SourceAspect:  # narrower
            ScaleFactor = TargetSize[0] / SourceSize[0]
            GetZoomFactor.Cached_Result = (
                ScaleFactor,
                0,
                (TargetSize[1] - TargetSize[0] / SourceAspect) / 2,
            )
        elif TargetAspect > SourceAspect:  # wider
            ScaleFactor = TargetSize[1] / SourceSize[1]
            GetZoomFactor.Cached_Result = (
                ScaleFactor,
                (TargetSize[0] - TargetSize[1] * SourceAspect) / 2,
                0,
            )
        else:
            GetZoomFactor.Cached_Result = (TargetSize[0] / SourceSize[0], 0, 0)
        return GetZoomFactor.Cached_Result
    except ZeroDivisionError:
        GetZoomFactor.Cached_Result = (1, 0, 0)
        return GetZoomFactor.Cached_Result


# Calculation is based on https://github.com/jabbany/CommentCoreLibrary/issues/5#issuecomment-40087282
#                     and https://github.com/m13253/danmaku2ass/issues/7#issuecomment-41489422
# ASS FOV = width*4/3.0
# But Flash FOV = width/math.tan(100*math.pi/360.0)/2 will be used instead
# Result: (transX, transY, rotX, rotY, rotZ, scaleX, scaleY)
def ConvertFlashRotation(rotY, rotZ, X, Y, width, height):
    def WrapAngle(deg):
        return 180 - ((180 - deg) % 360)

    rotY = WrapAngle(rotY)
    rotZ = WrapAngle(rotZ)
    if rotY in (90, -90):
        rotY -= 1
    if rotY == 0 or rotZ == 0:
        outX = 0
        outY = -rotY  # Positive value means clockwise in Flash
        outZ = -rotZ
        rotY *= math.pi / 180.0
        rotZ *= math.pi / 180.0
    else:
        rotY *= math.pi / 180.0
        rotZ *= math.pi / 180.0
        outY = (
            math.atan2(-math.sin(rotY) * math.cos(rotZ), math.cos(rotY)) * 180 / math.pi
        )
        outZ = (
            math.atan2(-math.cos(rotY) * math.sin(rotZ), math.cos(rotZ)) * 180 / math.pi
        )
        outX = math.asin(math.sin(rotY) * math.sin(rotZ)) * 180 / math.pi
    trX = (
        (X * math.cos(rotZ) + Y * math.sin(rotZ)) / math.cos(rotY)
        + (1 - math.cos(rotZ) / math.cos(rotY)) * width / 2
        - math.sin(rotZ) / math.cos(rotY) * height / 2
    )
    trY = (
        Y * math.cos(rotZ)
        - X * math.sin(rotZ)
        + math.sin(rotZ) * width / 2
        + (1 - math.cos(rotZ)) * height / 2
    )
    trZ = (trX - width / 2) * math.sin(rotY)
    FOV = width * math.tan(2 * math.pi / 9.0) / 2
    try:
        scaleXY = FOV / (FOV + trZ)
    except ZeroDivisionError:
        logging.error("Rotation makes object behind the camera: trZ == %.0f" % trZ)
        scaleXY = 1
    trX = (trX - width / 2) * scaleXY + width / 2
    trY = (trY - height / 2) * scaleXY + height / 2
    if scaleXY < 0:
        scaleXY = -scaleXY
        outX += 180
        outY += 180
        logging.error(
            "Rotation makes object behind the camera: trZ == %.0f < %.0f" % (trZ, FOV)
        )
    return (
        trX,
        trY,
        WrapAngle(outX),
        WrapAngle(outY),
        WrapAngle(outZ),
        scaleXY * 100,
        scaleXY * 100,
    )


def ProcessComments(
    comments,
    f,
    width,
    height,
    bottomReserved,
    fontface,
    fontsize,
    alpha,
    duration_marquee,
    duration_still,
    filters_regex,
    reduced,
    progress_callback,
):
    styleid = "Danmaku2ASS_%04x" % random.randint(0, 0xFFFF)
    WriteASSHead(f, width, height, fontface, fontsize, alpha, styleid)
    rows = [[None] * (height - bottomReserved + 1) for i in range(4)]
    for idx, i in enumerate(comments):
        if progress_callback and idx % 1000 == 0:
            progress_callback(idx, len(comments))
        if isinstance(i[4], int):
            skip = False
            for filter_regex in filters_regex:
                if filter_regex and filter_regex.search(i[3]):
                    skip = True
                    break
            if skip:
                continue
            row = 0
            rowmax = height - bottomReserved - i[7]
            while row <= rowmax:
                freerows = TestFreeRows(
                    rows,
                    i,
                    row,
                    width,
                    height,
                    bottomReserved,
                    duration_marquee,
                    duration_still,
                )
                if freerows >= i[7]:
                    MarkCommentRow(rows, i, row)
                    WriteComment(
                        f,
                        i,
                        row,
                        width,
                        height,
                        bottomReserved,
                        fontsize,
                        duration_marquee,
                        duration_still,
                        styleid,
                    )
                    break
                else:
                    row += freerows or 1
            else:
                if not reduced:
                    row = FindAlternativeRow(rows, i, height, bottomReserved)
                    MarkCommentRow(rows, i, row)
                    WriteComment(
                        f,
                        i,
                        row,
                        width,
                        height,
                        bottomReserved,
                        fontsize,
                        duration_marquee,
                        duration_still,
                        styleid,
                    )
        elif i[4] == "bilipos":
            WriteCommentBilibiliPositioned(f, i, width, height, styleid)
        elif i[4] == "acfunpos":
            WriteCommentAcfunPositioned(f, i, width, height, styleid)
        else:
            logging.warning(_("Invalid comment: %r") % i[3])
    if progress_callback:
        progress_callback(len(comments), len(comments))


def TestFreeRows(
    rows, c, row, width, height, bottomReserved, duration_marquee, duration_still
):
    res = 0
    rowmax = height - bottomReserved
    targetRow = None
    if c[4] in (1, 2):
        while row < rowmax and res < c[7]:
            if targetRow != rows[c[4]][row]:
                targetRow = rows[c[4]][row]
                if targetRow and targetRow[0] + duration_still > c[0]:
                    break
            row += 1
            res += 1
    else:
        try:
            thresholdTime = c[0] - duration_marquee * (1 - width / (c[8] + width))
        except ZeroDivisionError:
            thresholdTime = c[0] - duration_marquee
        while row < rowmax and res < c[7]:
            if targetRow != rows[c[4]][row]:
                targetRow = rows[c[4]][row]
                try:
                    if targetRow and (
                        targetRow[0] > thresholdTime
                        or targetRow[0]
                        + targetRow[8] * duration_marquee / (targetRow[8] + width)
                        > c[0]
                    ):
                        break
                except ZeroDivisionError:
                    pass
            row += 1
            res += 1
    return res


def FindAlternativeRow(rows, c, height, bottomReserved):
    res = 0
    for row in range(height - bottomReserved - math.ceil(c[7])):
        if not rows[c[4]][row]:
            return row
        elif rows[c[4]][row][0] < rows[c[4]][res][0]:
            res = row
    return res


def MarkCommentRow(rows, c, row):
    try:
        for i in range(row, row + math.ceil(c[7])):
            rows[c[4]][i] = c
    except IndexError:
        pass


def WriteASSHead(f, width, height, fontface, fontsize, alpha, styleid):
    f.write(
        """[Script Info]
; Script generated by Danmaku2ASS
; https://github.com/m13253/danmaku2ass
Script Updated By: Danmaku2ASS (https://github.com/m13253/danmaku2ass)
ScriptType: v4.00+
PlayResX: %(width)d
PlayResY: %(height)d
Aspect Ratio: %(width)d:%(height)d
Collisions: Normal
WrapStyle: 2
ScaledBorderAndShadow: yes
YCbCr Matrix: TV.601
[V4+ Styles]
Format: Name, Fontname, Fontsize, PrimaryColour, SecondaryColour, OutlineColour, BackColour, Bold, Italic, Underline, StrikeOut, ScaleX, ScaleY, Spacing, Angle, BorderStyle, Outline, Shadow, Alignment, MarginL, MarginR, MarginV, Encoding
Style: %(styleid)s, %(fontface)s, %(fontsize).0f, &H%(alpha)02XFFFFFF, &H%(alpha)02XFFFFFF, &H%(alpha)02X000000, &H%(alpha)02X000000, 0, 0, 0, 0, 100, 100, 0.00, 0.00, 1, %(outline).0f, 0, 7, 0, 0, 0, 0
[Events]
Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
"""
        % {
            "width": width,
            "height": height,
            "fontface": fontface,
            "fontsize": fontsize,
            "alpha": 255 - round(alpha * 255),
            "outline": max(fontsize / 25.0, 1),
            "styleid": styleid,
        }
    )


def WriteComment(
    f,
    c,
    row,
    width,
    height,
    bottomReserved,
    fontsize,
    duration_marquee,
    duration_still,
    styleid,
):
    text = ASSEscape(c[3])
    styles = []
    if c[4] == 1:
        styles.append(
            "\\an8\\pos(%(halfwidth)d, %(row)d)" % {"halfwidth": width / 2, "row": row}
        )
        duration = duration_still
    elif c[4] == 2:
        styles.append(
            "\\an2\\pos(%(halfwidth)d, %(row)d)"
            % {"halfwidth": width / 2, "row": ConvertType2(row, height, bottomReserved)}
        )
        duration = duration_still
    elif c[4] == 3:
        styles.append(
            "\\move(%(neglen)d, %(row)d, %(width)d, %(row)d)"
            % {"width": width, "row": row, "neglen": -math.ceil(c[8])}
        )
        duration = duration_marquee
    else:
        styles.append(
            "\\move(%(width)d, %(row)d, %(neglen)d, %(row)d)"
            % {"width": width, "row": row, "neglen": -math.ceil(c[8])}
        )
        duration = duration_marquee
    if not (-1 < c[6] - fontsize < 1):
        styles.append("\\fs%.0f" % c[6])
    if c[5] != 0xFFFFFF:
        styles.append("\\c&H%s&" % ConvertColor(c[5]))
        if c[5] == 0x000000:
            styles.append("\\3c&HFFFFFF&")
    f.write(
        "Dialogue: 2,%(start)s,%(end)s,%(styleid)s,,0000,0000,0000,,{%(styles)s}%(text)s\n"
        % {
            "start": ConvertTimestamp(c[0]),
            "end": ConvertTimestamp(c[0] + duration),
            "styles": "".join(styles),
            "text": text,
            "styleid": styleid,
        }
    )


def ASSEscape(s):
    def ReplaceLeadingSpace(s):
        sstrip = s.strip(" ")
        slen = len(s)
        if slen == len(sstrip):
            return s
        else:
            llen = slen - len(s.lstrip(" "))
            rlen = slen - len(s.rstrip(" "))
            return "".join(("\u2007" * llen, sstrip, "\u2007" * rlen))

    return "\\N".join(
        (
            ReplaceLeadingSpace(i) or " "
            for i in str(s)
            .replace("\\", "\\\\")
            .replace("{", "\\{")
            .replace("}", "\\}")
            .split("\n")
        )
    )


def CalculateLength(s):
    return max(map(len, s.split("\n")))  # May not be accurate


def ConvertTimestamp(timestamp):
    timestamp = round(timestamp * 100.0)
    hour, minute = divmod(timestamp, 360000)
    minute, second = divmod(minute, 6000)
    second, centsecond = divmod(second, 100)
    return "%d:%02d:%02d.%02d" % (int(hour), int(minute), int(second), int(centsecond))


def ConvertColor(RGB, width=1280, height=576):
    if RGB == 0x000000:
        return "000000"
    elif RGB == 0xFFFFFF:
        return "FFFFFF"
    R = (RGB >> 16) & 0xFF
    G = (RGB >> 8) & 0xFF
    B = RGB & 0xFF
    if width < 1280 and height < 576:
        return "%02X%02X%02X" % (B, G, R)
    else:  # VobSub always uses BT.601 colorspace, convert to BT.709
        ClipByte = lambda x: 255 if x > 255 else 0 if x < 0 else round(x)
        return "%02X%02X%02X" % (
            ClipByte(
                R * 0.00956384088080656
                + G * 0.03217254540203729
                + B * 0.95826361371715607
            ),
            ClipByte(
                R * -0.10493933142075390
                + G * 1.17231478191855154
                + B * -0.06737545049779757
            ),
            ClipByte(
                R * 0.91348912373987645
                + G * 0.07858536372532510
                + B * 0.00792551253479842
            ),
        )


def ConvertType2(row, height, bottomReserved):
    return height - bottomReserved - row


def ConvertToFile(filename_or_file, *args, **kwargs):
    if isinstance(filename_or_file, bytes):
        filename_or_file = str(bytes(filename_or_file).decode("utf-8", "replace"))
    if isinstance(filename_or_file, str):
        return open(filename_or_file, *args, **kwargs)
    else:
        return filename_or_file


def FilterBadChars(f):
    s = f.read()
    s = re.sub("[\\x00-\\x08\\x0b\\x0c\\x0e-\\x1f]", "\ufffd", s)
    return io.StringIO(s)


class safe_list(list):
    def get(self, index, default=None):
        try:
            return self[index]
        except IndexError:
            return default


def export(func):
    global __all__
    try:
        __all__.append(func.__name__)
    except NameError:
        __all__ = [func.__name__]
    return func


@export
def Danmaku2ASS(
    input_files,
    input_format,
    output_file,
    stage_width,
    stage_height,
    reserve_blank=0,
    font_face=_("(FONT) sans-serif")[7:],
    font_size=25.0,
    text_opacity=1.0,
    duration_marquee=5.0,
    duration_still=5.0,
    comment_filter=None,
    comment_filters_file=None,
    is_reduce_comments=False,
    progress_callback=None,
):
    comment_filters = [comment_filter]
    if comment_filters_file:
        with open(comment_filters_file, "r") as f:
            d = f.readlines()
            comment_filters.extend([i.strip() for i in d])
    filters_regex = []
    for comment_filter in comment_filters:
        try:
            if comment_filter:
                filters_regex.append(re.compile(comment_filter))
        except:
            raise ValueError(_("Invalid regular expression: %s") % comment_filter)
    fo = None
    comments = ReadComments(input_files, input_format, font_size)
    try:
        if output_file:
            fo = ConvertToFile(
                output_file, "w", encoding="utf-8-sig", errors="replace", newline="\r\n"
            )
        else:
            fo = sys.stdout
        ProcessComments(
            comments,
            fo,
            stage_width,
            stage_height,
            reserve_blank,
            font_face,
            font_size,
            text_opacity,
            duration_marquee,
            duration_still,
            filters_regex,
            is_reduce_comments,
            progress_callback,
        )
    finally:
        if output_file and fo != output_file:
            fo.close()


@export
def ReadComments(input_files, input_format, font_size=25.0, progress_callback=None):
    if isinstance(input_files, bytes):
        input_files = str(bytes(input_files).decode("utf-8", "replace"))
    if isinstance(input_files, str):
        input_files = [input_files]
    else:
        input_files = list(input_files)
    comments = []
    for idx, i in enumerate(input_files):
        if progress_callback:
            progress_callback(idx, len(input_files))
        with ConvertToFile(i, "r", encoding="utf-8", errors="replace") as f:
            s = f.read()
            str_io = io.StringIO(s)
            if input_format == "autodetect":
                CommentProcessor = GetCommentProcessor(str_io)
                if not CommentProcessor:
                    raise ValueError(_("Failed to detect comment file format: %s") % i)
            else:
                CommentProcessor = CommentFormatMap.get(input_format)
                if not CommentProcessor:
                    raise ValueError(
                        _("Unknown comment file format: %s") % input_format
                    )
            comments.extend(CommentProcessor(FilterBadChars(str_io), font_size))
    if progress_callback:
        progress_callback(len(input_files), len(input_files))
    comments.sort()
    return comments


@export
def GetCommentProcessor(input_file):
    return CommentFormatMap.get(ProbeCommentFormat(input_file))


def main():
    logging.basicConfig(format="%(levelname)s: %(message)s")
    if len(sys.argv) == 1:
        sys.argv.append("--help")
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f",
        "--format",
        metavar=_("FORMAT"),
        help=_("Format of input file (autodetect|%s) [default: autodetect]")
        % "|".join(i for i in CommentFormatMap),
        default="autodetect",
    )
    parser.add_argument("-o", "--output", metavar=_("OUTPUT"), help=_("Output file"))
    parser.add_argument(
        "-s",
        "--size",
        metavar=_("WIDTHxHEIGHT"),
        required=True,
        help=_("Stage size in pixels"),
    )
    parser.add_argument(
        "-fn",
        "--font",
        metavar=_("FONT"),
        help=_("Specify font face [default: %s]") % _("(FONT) sans-serif")[7:],
        default=_("(FONT) sans-serif")[7:],
    )
    parser.add_argument(
        "-fs",
        "--fontsize",
        metavar=_("SIZE"),
        help=(_("Default font size [default: %s]") % 25),
        type=float,
        default=25.0,
    )
    parser.add_argument(
        "-a",
        "--alpha",
        metavar=_("ALPHA"),
        help=_("Text opacity"),
        type=float,
        default=1.0,
    )
    parser.add_argument(
        "-dm",
        "--duration-marquee",
        metavar=_("SECONDS"),
        help=_("Duration of scrolling comment display [default: %s]") % 5,
        type=float,
        default=5.0,
    )
    parser.add_argument(
        "-ds",
        "--duration-still",
        metavar=_("SECONDS"),
        help=_("Duration of still comment display [default: %s]") % 5,
        type=float,
        default=5.0,
    )
    parser.add_argument(
        "-fl", "--filter", help=_("Regular expression to filter comments")
    )
    parser.add_argument(
        "-flf",
        "--filter-file",
        help=_("Regular expressions from file (one line one regex) to filter comments"),
    )
    parser.add_argument(
        "-p",
        "--protect",
        metavar=_("HEIGHT"),
        help=_("Reserve blank on the bottom of the stage"),
        type=int,
        default=0,
    )
    parser.add_argument(
        "-r",
        "--reduce",
        action="store_true",
        help=_("Reduce the amount of comments if stage is full"),
    )
    parser.add_argument(
        "file", metavar=_("FILE"), nargs="+", help=_("Comment file to be processed")
    )
    args = parser.parse_args()
    try:
        width, height = str(args.size).split("x", 1)
        width = int(width)
        height = int(height)
    except ValueError:
        raise ValueError(_("Invalid stage size: %r") % args.size)
    Danmaku2ASS(
        args.file,
        args.format,
        args.output,
        width,
        height,
        args.protect,
        args.font,
        args.fontsize,
        args.alpha,
        args.duration_marquee,
        args.duration_still,
        args.filter,
        args.filter_file,
        args.reduce,
    )


if __name__ == "__main__":
    main()



"""
bilibili_api
"""

import asyncio
import platform

# Modules
from . import (
    app,
    article,
    ass,
    audio,
    bangumi,
    black_room,
    channel,
    cheese,
    comment,
    dynamic,
    favorite_list,
    homepage,
    interactive_video,
    live,
    login_func,
    login,
    rank,
    search,
    settings,
    user,
    video_uploader,
    video,
)

# CONSTANTS
from .video import AUDIO_QUALITY, VIDEO_CODECS, VIDEO_QUALITY
from .article import ARTICLE_COLOR_MAP
from .black_room import BLACK_TYPE

HEADERS = {"User-Agent": "Mozilla/5.0", "Referer": "https://www.bilibili.com"}

# Functions
from .utils.parse_link import parse_link
from .utils.short import get_real_url
from .utils.sync import sync
from .utils.aid_bvid_transformer import aid2bvid, bvid2aid

# Models
from .utils.Credential import Credential
from .utils.parse_link import ResourceType
from .utils.Danmaku import Danmaku, DmMode, DmFontSize, SpecialDanmaku

# Errors
from .errors import *

# 如果系统为 Windows，则修改默认策略，以解决代理报错问题
if "windows" in platform.system().lower():
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
del asyncio, platform

# ALL
__all__ = [
    "app",
    "article",
    "ass",
    "audio",
    "bangumi",
    "black_room",
    "channel",
    "cheese",
    "comment",
    "dynamic",
    "favorite_list",
    "homepage",
    "interactive_video",
    "live",
    "login_func",
    "login",
    "rank",
    "search",
    "settings",
    "user",
    "video_uploader",
    "video",
    "ARTICLE_COLOR_MAP",
    "AUDIO_QUALITY",
    "ApiException",
    "ArgsException",
    "BLACK_TYPE",
    "Credential",
    "CredentialNoBiliJctException",
    "CredentialNoBuvid3Exception",
    "CredentialNoDedeUserIDException",
    "CredentialNoSessdataException",
    "Danmaku",
    "DanmakuClosedException",
    "DmFontSize",
    "DmMode",
    "HEADERS",
    "LoginError",
    "NetworkException",
    "ResourceType",
    "ResponseCodeException",
    "ResponseException",
    "SpecialDanmaku",
    "VIDEO_CODECS",
    "VIDEO_QUALITY",
    "VideoUploadException",
    "aid2bvid",
    "bvid2aid",
    "get_real_url",
    "parse_link",
    "sync",
]



"""
bilibili_api.black_room

小黑屋
"""

from enum import Enum

from .utils.utils import get_api
from .utils.network_httpx import request
from .utils.Credential import Credential

# 违规类型代码
BLACK_TYPE = {
    0: "全部",
    1: "刷屏",
    2: "抢楼",
    3: "发布色情低俗信息",
    4: "发布赌博诈骗信息",
    5: "发布违禁相关信息",
    6: "发布垃圾广告信息",
    7: "发布人身攻击言论",
    8: "发布侵犯他人隐私信息",
    9: "发布引战言论",
    10: "发布剧透信息",
    11: "恶意添加无关标签",
    12: "恶意删除他人标签",
    13: "发布色情信息",
    14: "发布低俗信息",
    15: "发布暴力血腥信息",
    16: "涉及恶意投稿行为",
    17: "发布非法网站信息",
    18: "发布传播不实信息",
    19: "发布怂恿教唆信息",
    20: "恶意刷屏",
    21: "账号违规",
    22: "恶意抄袭",
    23: "冒充自制原创",
    24: "发布青少年不良内容",
    25: "破坏网络安全",
    26: "发布虚假误导信息",
    27: "仿冒官方认证账号",
    28: "发布不适宜内容",
    29: "违反运营规则",
    30: "恶意创建话题",
    31: "发布违规抽奖",
    32: "恶意冒充他人",
}


class BlackFrom(Enum):
    """
    违规来源

    SYSTEM: 系统封禁
    ADMIN: 风纪仲裁
    ALL: 全部
    """

    SYSTEM = 0
    ADMIN = 1
    ALL = None


API = get_api("black-room")


async def get_blocked_list(
    from_: BlackFrom = BlackFrom.ALL,
    type_: int = 0,
    pn: int = 1,
    credential: Credential = None,
):
    """
    获取小黑屋中的违规列表

    Args:
        from_(BlackFrom)      : 违规来源. Defaults to BlackFrom.ALL
        type_(int)            : 违规类型. 查看 black_room.BLACK_TYPE。Defaults to 0 (ALL)
        pn(int)               : 页数. Defaults to 1
        credential(Credential): 凭据, Defaults to None
    """
    credential = credential if credential else Credential()
    api = API["info"]
    params = {"pn": pn, "otype": type_}
    if from_.value != None:
        params["btype"] = from_.value
    return await request("GET", api["url"], params=params, credential=credential)


class BlackRoom:
    def __init__(self, black_room_id: int, credential: Credential = None):
        """
        Args:
            black_room_id(int)    : 小黑屋 id
            credential(Credential): 凭据类
        """
        self.__id = black_room_id
        self.credential = credential if credential else Credential()

    async def get_details(self):
        """
        获取小黑屋详细信息
        """
        api = API["detail"]
        params = {"id": self.__id}
        return await request(
            "GET", api["url"], params=params, credential=self.credential
        )

    async def get_id(self):
        """
        获取小黑屋 id
        """
        return self.__id



# Bibibili API Documentions

import webbrowser
import sys

try:
    from PyQt5 import QtCore, QtWidgets, QtGui, QtWebEngineWidgets
except:
    PYQT5 = False
else:
    PYQT5 = True


def main():
    if not PYQT5:
        webbrowser.open("https://nemo2011.github.io/bilibili-api")
    else:
        QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling)
        app = QtWidgets.QApplication(sys.argv)
        mainwindow = QtWidgets.QMainWindow()
        mainwindow.resize(800, 600)
        webengine = QtWebEngineWidgets.QWebEngineView(mainwindow)
        webengine.setGeometry(QtCore.QRect(0, 0, 800, 600))
        webengine.setUrl(QtCore.QUrl("https://nemo2011.github.io/bilibili-api"))

        def timerEvent(*args, **kwargs):
            webengine.setGeometry(
                QtCore.QRect(0, 0, mainwindow.width(), mainwindow.height())
            )

        mainwindow.timerEvent = timerEvent
        mainwindow.startTimer(1)
        mainwindow.show()
        sys.exit(app.exec_())


if __name__ == "__main__":
    main()



import setuptools

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("requirements.txt", "r", encoding="utf8") as f:
    requires = f.read()

setuptools.setup(
    name="bilibili-api-python",
    version="12.4.0",
    license="GPLv3+",
    author="MoyuScript, Nemo2011",
    description="哔哩哔哩的各种 API 调用便捷整合（视频、动态、直播等），另外附加一些常用的功能。",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=[
        "bilibili_api",
        "bilibili_api.utils",
        "bilibili_api.exceptions",
        "bilibili_api.errors", 
        "bilibili_api._pyinstaller",
    ],
    keywords=["bilibili", "api", "spider"],
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Natural Language :: Chinese (Simplified)",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    package_data={"": ["data/**/*.*", "requirements.txt", "data/*.*", "html/*.*"]},
    install_requires=requires.splitlines(),
    url="https://github.com/nemo2011/bilibili-api",
    python_requires=">=3.8",
    entry_points={
        "pyinstaller40": [
            "hook-dirs = bilibili_api._pyinstaller.entry_points:get_hook_dirs"
        ],
        "console_scripts": ["bilibili-api = bilibili_api.__main__:main"],
    },
)



import asyncio
import datetime
import json
import logging
import time
from dataclasses import dataclass
from typing import Union

import httpx
from apscheduler.schedulers.asyncio import AsyncIOScheduler

from .user import get_self_info
from .utils.AsyncEvent import AsyncEvent
from .utils.Credential import Credential
from .utils.network_httpx import request
from .utils.utils import get_api
from .video import Video

API = get_api("session")


async def fetch_session_msgs(
    talker_id: int, credential: Credential, session_type: int = 1, begin_seqno: int = 0
):
    """
    获取指定用户的近三十条消息

    Args:
        talker_id    (int)       : 用户 UID
        credential   (Credential): Credential
        session_type (int)       : 会话类型 1 私聊 2 应援团
        begin_seqno  (int)       : 起始 Seqno

    Returns:
        调用 API 返回结果
    """

    credential.raise_for_no_sessdata()
    params = {
        "talker_id": talker_id,
        "session_type": session_type,
        "begin_seqno": begin_seqno,
    }
    api = API["session"]["fetch"]

    return await request("GET", api["url"], params=params, credential=credential)


async def new_sessions(
    credential: Credential, begin_ts: int = int(time.time() * 1000000)
):
    """
    获取新消息

    Args:
        credential (Credential): Credential
        begin_ts   (int)       : 起始时间戳

    Returns:
        调用 API 返回结果
    """

    credential.raise_for_no_sessdata()
    params = {"begin_ts": begin_ts, "build": 0, "mobi_app": "web"}
    api = API["session"]["new"]

    return await request("GET", api["url"], params=params, credential=credential)


async def get_sessions(credential: Credential, session_type: int = 4):
    """
    获取已有消息

    Args:
        credential   (Credential): Credential
        session_type (int)       : 会话类型 1: 私聊, 2: 通知, 3: 应援团, 4: 全部

    Returns:
        调用 API 返回结果
    """

    credential.raise_for_no_sessdata()
    params = {
        "session_type": session_type,
        "group_fold": 1,
        "unfollow_fold": 0,
        "sort_rule": 2,
        "build": 0,
        "mobi_app": "web",
    }
    api = API["session"]["get"]

    return await request("GET", api["url"], params=params, credential=credential)


async def send_msg(credential: Credential, receiver_id: int, text: str):
    """
    给用户发送私聊信息。目前仅支持纯文本。

    Args:
        credential (Credential): 凭证
        receiver_id (int): 接收者 UID
        text (str): 信息内容。

    Returns:
        dict: 调用接口返回的内容。
    """
    credential.raise_for_no_sessdata()
    credential.raise_for_no_bili_jct()

    api = API["operate"]["send_msg"]
    self_info = await get_self_info(credential)
    sender_uid = self_info["mid"]

    data = {
        "msg[sender_uid]": sender_uid,
        "msg[receiver_id]": receiver_id,
        "msg[receiver_type]": 1,
        "msg[msg_type]": 1,
        "msg[msg_status]": 0,
        "msg[content]": json.dumps({"content": text}),
        "msg[dev_id]": "B9A37BF3-AA9D-4076-A4D3-366AC8C4C5DB",
        "msg[new_face_version]": "0",
        "msg[timestamp]": int(time.time()),
        "from_filework": 0,
        "build": 0,
        "mobi_app": "web",
    }
    return await request("POST", url=api["url"], data=data, credential=credential)


@dataclass
class Picture:
    """
    图片类，包含图片链接、尺寸以及下载操作。
    """

    height: int
    imageType: str
    original: int
    size: str
    url: str
    width: int

    async def download(self, filepath: str = ""):
        if not filepath:
            filepath = self.url.replace("https://message.biliimg.com/bfs/im/", "")
        async with httpx.AsyncClient() as session:
            r = await session.get(self.url)
            with open(filepath, "wb") as fp:
                fp.write(r.content)


class Event:
    """
    事件参数:
    + receiver_id:   收信人 UID
    + receiver_type: 收信人类型，1: 私聊, 2: 应援团通知, 3: 应援团
    + sender_uid:    发送人 UID
    + talker_id:     对话人 UID
    + msg_seqno:     事件 Seqno
    + msg_type:      事件类型
    + msg_key:       事件唯一编号
    + timestamp:     事件时间戳
    + content:       事件内容

    事件类型:
    + TEXT:           纯文字消息
    + PICTURE:        图片消息
    + WITHDRAW:       撤回消息
    + GROUPS_PICTURE: 应援团图片，但似乎不常触发，一般使用 PICTURE 即可
    + SHARE_VIDEO:    分享视频
    + NOTICE:         系统通知
    + PUSHED_VIDEO:   UP主推送的视频
    + WELCOME:        新成员加入应援团欢迎
    """

    receiver_id: int
    receiver_type: int
    sender_uid: int
    talker_id: int
    msg_seqno: int
    msg_type: int
    msg_key: int
    timestamp: int
    content: Union[str, int, Picture, Video]

    TEXT = "1"
    PICTURE = "2"
    WITHDRAW = "5"
    GROUPS_PICTURE = "6"
    SHARE_VIDEO = "7"
    NOTICE = "10"
    PUSHED_VIDEO = "11"
    WELCOME = "306"

    def __init__(self, data: dict, self_uid: int):
        """
        信息事件类型:
            data: 接收到的事件详细信息
            self_uid: 用户自身 UID
        """
        self.__dict__.update(data)
        self.uid = self_uid

        try:
            self.__content()
        except AttributeError:
            print(data)

    def __str__(self):
        if self.receiver_type == 1:
            if self.receiver_id == self.uid:
                msg_type = "收到"
                user_id = self.sender_uid
            elif self.sender_uid == self.uid:
                msg_type = "发送"
                user_id = self.receiver_id

        elif self.receiver_type == 2:
            user_id = self.receiver_id
            if self.sender_uid == self.uid:
                msg_type = "发送应援团"
            elif self.sender_uid == 0:
                msg_type = "系统提示"
            else:
                msg_type = "收到应援团"

        return f"{msg_type} {user_id} 信息 {self.content}({self.timestamp})"

    def __content(self):
        """更新消息内容"""

        content: dict = json.loads(self.content)
        mt = str(self.msg_type)

        if mt == Event.TEXT:
            self.content = content.get("content")

        elif mt == Event.WELCOME:
            self.content = content.get("content") + str(content.get("group_id"))

        elif mt == Event.WITHDRAW:
            self.content = str(content)

        elif mt == Event.PICTURE or mt == Event.GROUPS_PICTURE:
            self.content = Picture(**content)

        elif mt == Event.SHARE_VIDEO or mt == Event.PUSHED_VIDEO:
            self.content = Video(bvid=content.get("bvid"), aid=content.get("id"))

        elif mt == Event.NOTICE:
            self.content = content["title"] + " " + content["text"]

        else:
            print(mt, content)


class Session(AsyncEvent):
    """
    会话类，用来开启消息监听。
    """

    def __init__(self, credential: Credential, debug=False):
        super().__init__()
        # 会话状态
        self.__status = 0

        # 已获取会话中最大的时间戳 默认当前时间
        self.maxTs = int(time.time() * 1000000)

        # 会话UID为键 会话中最大Seqno为值
        self.maxSeqno = dict()

        # 凭证
        self.credential = credential

        # 异步定时任务框架
        self.sched = AsyncIOScheduler(timezone="Asia/Shanghai")

        # 已接收的所有事件 用于撤回时找回
        self.events = dict()

        # logging
        self.logger = logging.getLogger("Session")
        self.logger.setLevel(logging.DEBUG if debug else logging.INFO)
        if not self.logger.handlers:
            handler = logging.StreamHandler()
            handler.setFormatter(
                logging.Formatter(
                    "[%(asctime)s][%(levelname)s]: %(message)s", "%Y-%m-%d %H:%M:%S"
                )
            )
            self.logger.addHandler(handler)

    def get_status(self):
        """
        获取连接状态

        Returns:
            int: 0 初始化，1 已连接，2 断开连接中，3 已断开，4 错误
        """
        return self.__status

    async def run(self, except_self: bool = True):
        """
        非阻塞异步爬虫 定时发送请求获取消息

        Args:
            except_self: bool 是否排除自己发出的消息，默认是
        """

        # 获取自身UID 用于后续判断消息是发送还是接收
        self_info = await get_self_info(self.credential)
        self.uid = self_info["mid"]

        # 初始化 只接收开始运行后的新消息
        js = await get_sessions(self.credential)
        self.maxSeqno = {
            _session["talker_id"]: _session["max_seqno"]
            for _session in js.get("session_list", [])
        }

        # 间隔 6 秒轮询消息列表 之前设置 3 秒询一次 跑了一小时给我账号冻结了
        @self.sched.scheduled_job(
            "interval",
            id="query",
            seconds=6,
            max_instances=3,
            next_run_time=datetime.datetime.now(),
        )
        async def qurey():
            js: dict = await new_sessions(self.credential, self.maxTs)
            if js.get("session_list") is None:
                return

            pending = set()
            for session in js["session_list"]:
                self.maxTs = max(self.maxTs, session["session_ts"])
                pending.add(
                    asyncio.create_task(
                        fetch_session_msgs(
                            session["talker_id"],
                            self.credential,
                            session["session_type"],
                            self.maxSeqno.get(session["talker_id"]),
                        )
                    )
                )
                self.maxSeqno[session["talker_id"]] = session["max_seqno"]

            while pending:
                done, pending = await asyncio.wait(pending)
                for done_task in done:
                    result: dict = await done_task
                    if result is None or result.get("messages") is None:
                        continue
                    for message in result.get("messages", [])[::-1]:
                        event = Event(message, self.uid)
                        if str(event.msg_type) == Event.WITHDRAW:
                            self.logger.info(
                                str(
                                    self.events.get(
                                        event.content, f"key={event.content}"
                                    )
                                )
                                + f" 被撤回({event.timestamp})"
                            )
                        else:
                            self.logger.info(event)

                        # 自己发出的消息不派发任务
                        if event.sender_uid != self.uid or not except_self:
                            self.dispatch(str(event.msg_type), event)

                        self.events.update({str(event.msg_key): event})

            self.logger.debug(f"maxTs = {self.maxTs}")

        self.__status = 1
        self.sched.start()
        self.logger.info("开始轮询")

    async def start(self, except_self: bool = True):
        """
        阻塞异步启动 通过调用 self.close() 后可断开连接

        Args:
            except_self: bool 是否排除自己发出的消息，默认是
        """

        await self.run(except_self)
        while self.get_status() < 2:
            await asyncio.sleep(1)

        if self.get_status() == 2:
            self.__status = 3

    async def reply(self, event: Event, text: str):
        """
        快速回复消息

        Args:
            event: Event 要回复的消息
            text:  str   要回复的文字内容
        Returns:
            dict: 调用接口返回的内容。
        """

        if self.uid == event.sender_uid:
            self.logger.error("不能给自己发送消息哦~")
        else:
            return await send_msg(self.credential, event.sender_uid, text)

    def close(self):
        """结束轮询"""

        self.sched.remove_job("query")
        self.__status = 2
        self.logger.info("结束轮询")
